@@ -9,7 +9,7 @@ from textwrap import dedent, indent
 from typing import (  # no type comments
     TYPE_CHECKING,
     Any,
-    Callable,
+    Literal,
     NewType,
     Optional,
     TypeVar,
@@ -20,7 +20,7 @@ from typing import (  # no type comments
 import pytest
 
 if TYPE_CHECKING:
-    from collections.abc import AsyncGenerator
+    from collections.abc import AsyncGenerator, Callable
     from io import StringIO
     from mailbox import Mailbox
     from types import CodeType, ModuleType
@@ -31,6 +31,28 @@ T = TypeVar("T")
 W = NewType("W", str)
 
 
+@dataclass
+class WarningInfo:
+    """Properties and assertion methods for warnings."""
+
+    regexp: str
+    type: str
+
+    def assert_regexp(self, message: str) -> None:
+        regexp = self.regexp
+        msg = f"Regex pattern did not match.\n Regex: {regexp!r}\n Input: {message!r}"
+        assert re.search(regexp, message), msg
+
+    def assert_type(self, message: str) -> None:
+        expected = f"[{self.type}]"
+        msg = f"Warning did not contain type and subtype.\n Expected: {expected}\n Input: {message}"
+        assert expected in message, msg
+
+    def assert_warning(self, message: str) -> None:
+        self.assert_regexp(message)
+        self.assert_type(message)
+
+
 def expected(expected: str, **options: dict[str, Any]) -> Callable[[T], T]:
     def dec(val: T) -> T:
         val.EXPECTED = expected
@@ -40,9 +62,9 @@ def expected(expected: str, **options: dict[str, Any]) -> Callable[[T], T]:
     return dec
 
 
-def warns(pattern: str) -> Callable[[T], T]:
+def warns(info: WarningInfo) -> Callable[[T], T]:
     def dec(val: T) -> T:
-        val.WARNING = pattern
+        val.WARNING = info
         return val
 
     return dec
@@ -50,7 +72,7 @@ def warns(pattern: str) -> Callable[[T], T]:
 
 @expected("mod.get_local_function()")
 def get_local_function():  # noqa: ANN201
-    def wrapper(self) -> str:  # noqa: ANN001, ARG001
+    def wrapper(self) -> str:  # noqa: ANN001
         """
         Wrapper
         """
@@ -58,7 +80,7 @@ def get_local_function():  # noqa: ANN201
     return wrapper
 
 
-@warns("Cannot handle as a local function")
+@warns(WarningInfo(regexp="Cannot handle as a local function", type="sphinx_autodoc_typehints.local_function"))
 @expected(
     """\
 class mod.Class(x, y, z=None)
@@ -70,7 +92,7 @@ class mod.Class(x, y, z=None)
 
       * **y** ("int") -- bar
 
-      * **z** ("Optional"["str"]) -- baz
+      * **z** ("str" | "None") -- baz
 
    class InnerClass
 
@@ -95,7 +117,7 @@ class mod.Class(x, y, z=None)
 
          * **y** ("int") -- bar
 
-         * **z** ("Optional"["str"]) -- baz
+         * **z** ("str" | "None") -- baz
 
       Return type:
          "str"
@@ -109,7 +131,7 @@ class mod.Class(x, y, z=None)
 
          * **y** ("int") -- bar
 
-         * **z** ("Optional"["str"]) -- baz
+         * **z** ("str" | "None") -- baz
 
       Return type:
          "str"
@@ -127,7 +149,7 @@ class mod.Class(x, y, z=None)
 
          * **y** ("int") -- bar
 
-         * **z** ("Optional"["str"]) -- baz
+         * **z** ("str" | "None") -- baz
 
       Return type:
          "str"
@@ -149,10 +171,10 @@ class Class:
     :param z: baz
     """
 
-    def __init__(self, x: bool, y: int, z: Optional[str] = None) -> None:  # noqa: UP007
+    def __init__(self, x: bool, y: int, z: Optional[str] = None) -> None:
         pass
 
-    def a_method(self, x: bool, y: int, z: Optional[str] = None) -> str:  # noqa: UP007
+    def a_method(self, x: bool, y: int, z: Optional[str] = None) -> str:
         """
         Method docstring.
 
@@ -183,7 +205,7 @@ class Class:
         """
 
     @classmethod
-    def a_classmethod(cls, x: bool, y: int, z: Optional[str] = None) -> str:  # noqa: UP007
+    def a_classmethod(cls, x: bool, y: int, z: Optional[str] = None) -> str:
         """
         Classmethod docstring.
 
@@ -193,7 +215,7 @@ class Class:
         """
 
     @staticmethod
-    def a_staticmethod(x: bool, y: int, z: Optional[str] = None) -> str:  # noqa: UP007
+    def a_staticmethod(x: bool, y: int, z: Optional[str] = None) -> str:
         """
         Staticmethod docstring.
 
@@ -262,7 +284,7 @@ mod.function(x, y, z_=None)
 
       * **y** ("int") -- bar
 
-      * **z_** ("Optional"["str"]) -- baz
+      * **z_** ("str" | "None") -- baz
 
    Returns:
       something
@@ -271,7 +293,7 @@ mod.function(x, y, z_=None)
       bytes
 """,
 )
-def function(x: bool, y: int, z_: Optional[str] = None) -> str:  # noqa: ARG001, UP007
+def function(x: bool, y: int, z_: Optional[str] = None) -> str:
     """
     Function docstring.
 
@@ -299,7 +321,7 @@ mod.function_with_starred_documentation_param_names(*args, **kwargs)
       * ****kwargs** ("str") -- bar
 """,
 )
-def function_with_starred_documentation_param_names(*args: int, **kwargs: str):  # noqa: ANN201, ARG001
+def function_with_starred_documentation_param_names(*args: int, **kwargs: str):  # noqa: ANN201
     r"""
     Function docstring.
 
@@ -322,7 +344,7 @@ mod.function_with_escaped_default(x='\\\\x08')
       **x** ("str") -- foo
 """,
 )
-def function_with_escaped_default(x: str = "\b"):  # noqa: ANN201, ARG001
+def function_with_escaped_default(x: str = "\b"):  # noqa: ANN201
     """
     Function docstring.
 
@@ -330,7 +352,11 @@ def function_with_escaped_default(x: str = "\b"):  # noqa: ANN201, ARG001
     """
 
 
-@warns("Cannot resolve forward reference in type annotations")
+@warns(
+    WarningInfo(
+        regexp="Cannot resolve forward reference in type annotations", type="sphinx_autodoc_typehints.forward_reference"
+    )
+)
 @expected(
     """\
 mod.function_with_unresolvable_annotation(x)
@@ -341,7 +367,7 @@ mod.function_with_unresolvable_annotation(x)
       **x** (a.b.c) -- foo
 """,
 )
-def function_with_unresolvable_annotation(x: a.b.c):  # noqa: ANN201, ARG001, F821
+def function_with_unresolvable_annotation(x: a.b.c):  # noqa: ANN201, F821
     """
     Function docstring.
 
@@ -365,8 +391,8 @@ mod.function_with_typehint_comment(x, y)
 """,
 )
 def function_with_typehint_comment(  # noqa: ANN201
-    x,  # type: int  # noqa: ANN001, ARG001
-    y,  # type: str  # noqa: ANN001, ARG001
+    x,  # type: int  # noqa: ANN001
+    y,  # type: str  # noqa: ANN001
 ):
     # type: (...) -> None
     """
@@ -445,7 +471,7 @@ mod.function_with_typehint_comment_not_inline(x=None, *y, z, **kwargs)
    Function docstring.
 
    Parameters:
-      * **x** ("Union"["str", "bytes", "None"]) -- foo
+      * **x** ("str" | "bytes" | "None") -- foo
 
       * **y** ("str") -- bar
 
@@ -457,7 +483,7 @@ mod.function_with_typehint_comment_not_inline(x=None, *y, z, **kwargs)
       "None"
 """,
 )
-def function_with_typehint_comment_not_inline(x=None, *y, z, **kwargs):  # noqa: ANN001, ANN002, ANN003, ANN201, ARG001
+def function_with_typehint_comment_not_inline(x=None, *y, z, **kwargs):  # noqa: ANN001, ANN002, ANN003, ANN201
     # type: (Union[str, bytes, None], *str, bytes, **int) -> None
     """
     Function docstring.
@@ -476,7 +502,7 @@ class mod.ClassWithTypehintsNotInline(x=None)
    Class docstring.
 
    Parameters:
-      **x** ("Optional"["Callable"[["int", "bytes"], "int"]]) -- foo
+      **x** ("Callable"[["int", "bytes"], "int"] | "None") -- foo
 
    foo(x=1)
 
@@ -493,8 +519,7 @@ class mod.ClassWithTypehintsNotInline(x=None)
       Method docstring.
 
       Parameters:
-         **x** ("Optional"["Callable"[["int", "bytes"], "int"]]) --
-         foo
+         **x** ("Callable"[["int", "bytes"], "int"] | "None") -- foo
 
       Return type:
          "ClassWithTypehintsNotInline"
@@ -592,7 +617,7 @@ mod.mocked_import(x)
       **x** ("Mailbox") -- function
 """,
 )
-def mocked_import(x: Mailbox):  # noqa: ANN201, ARG001
+def mocked_import(x: Mailbox):  # noqa: ANN201
     """
     A docstring.
 
@@ -640,15 +665,15 @@ mod.func_with_overload(a, b)
    they must both have the same type.
 
    Parameters:
-      * **a** ("Union"["int", "str"]) -- The first thing
+      * **a** ("int" | "str") -- The first thing
 
-      * **b** ("Union"["int", "str"]) -- The second thing
+      * **b** ("int" | "str") -- The second thing
 
    Return type:
       "None"
 """,
 )
-def func_with_overload(a: Union[int, str], b: Union[int, str]) -> None:  # noqa: ARG001, UP007
+def func_with_overload(a: Union[int, str], b: Union[int, str]) -> None:
     """
     f does the thing. The arguments can either be ints or strings but they must
     both have the same type.
@@ -662,13 +687,66 @@ def func_with_overload(a: Union[int, str], b: Union[int, str]) -> None:  # noqa:
     """
 
 
+@expected(
+    """\
+mod.func_literals_long_format(a, b)
+
+   A docstring.
+
+   Parameters:
+      * **a** ("Literal"["'arg1'", "'arg2'"]) -- Argument that can
+        take either of two literal values.
+
+      * **b** ("Literal"["'arg1'", "'arg2'"]) -- Argument that can
+        take either of two literal values.
+
+   Return type:
+      "None"
+""",
+)
+def func_literals_long_format(a: Literal["arg1", "arg2"], b: Literal["arg1", "arg2"]) -> None:
+    """
+    A docstring.
+
+    :param a: Argument that can take either of two literal values.
+    :param b: Argument that can take either of two literal values.
+    """
+
+
+@expected(
+    """\
+mod.func_literals_short_format(a, b)
+
+   A docstring.
+
+   Parameters:
+      * **a** ("'arg1'" | "'arg2'") -- Argument that can take either
+        of two literal values.
+
+      * **b** ("'arg1'" | "'arg2'") -- Argument that can take either
+        of two literal values.
+
+   Return type:
+      "None"
+""",
+    python_display_short_literal_types=True,
+)
+def func_literals_short_format(a: Literal["arg1", "arg2"], b: Literal["arg1", "arg2"]) -> None:
+    """
+    A docstring.
+
+    :param a: Argument that can take either of two literal values.
+    :param b: Argument that can take either of two literal values.
+    """
+
+
 @expected(
     """\
 class mod.TestClassAttributeDocs
 
    A class
 
-   code: "Optional"["CodeType"]
+   code: "CodeType" | "None"
 
       An attribute
 """,
@@ -676,7 +754,7 @@ class mod.TestClassAttributeDocs
 class TestClassAttributeDocs:
     """A class"""
 
-    code: Optional[CodeType]  # noqa: UP007
+    code: Optional[CodeType]
     """An attribute"""
 
 
@@ -727,7 +805,7 @@ mod.func_with_parameters_and_stuff_after(a, b)
    More info about the function here.
 """,
 )
-def func_with_parameters_and_stuff_after(a: int, b: int) -> int:  # noqa: ARG001
+def func_with_parameters_and_stuff_after(a: int, b: int) -> int:
     """A func
 
     :param a: a tells us something
@@ -761,7 +839,7 @@ mod.func_with_rtype_in_weird_spot(a, b)
       int
 """,
 )
-def func_with_rtype_in_weird_spot(a: int, b: int) -> int:  # noqa: ARG001
+def func_with_rtype_in_weird_spot(a: int, b: int) -> int:
     """A func
 
     :param a: a tells us something
@@ -809,7 +887,7 @@ mod.empty_line_between_parameters(a, b)
    More stuff here.
 """,
 )
-def empty_line_between_parameters(a: int, b: int) -> int:  # noqa: ARG001
+def empty_line_between_parameters(a: int, b: int) -> int:
     """A func
 
     :param a: One of the following possibilities:
@@ -1002,7 +1080,7 @@ def napoleon_returns() -> CodeType:
 
     """,
 )
-def google_docstrings(arg1: CodeType, arg2: ModuleType) -> CodeType:  # noqa: ARG001
+def google_docstrings(arg1: CodeType, arg2: ModuleType) -> CodeType:
     """Summary line.
 
     Extended description of function.
@@ -1034,7 +1112,7 @@ def google_docstrings(arg1: CodeType, arg2: ModuleType) -> CodeType:  # noqa: AR
 
     """,
 )
-def docstring_with_multiline_note_after_params(param: int) -> None:  # noqa: ARG001
+def docstring_with_multiline_note_after_params(param: int) -> None:
     """Do something.
 
     Args:
@@ -1065,7 +1143,7 @@ def docstring_with_multiline_note_after_params(param: int) -> None:  # noqa: ARG
 
     """,
 )
-def docstring_with_bullet_list_after_params(param: int) -> None:  # noqa: ARG001
+def docstring_with_bullet_list_after_params(param: int) -> None:
     """Do something.
 
     Args:
@@ -1098,7 +1176,7 @@ def docstring_with_bullet_list_after_params(param: int) -> None:  # noqa: ARG001
 
     """,
 )
-def docstring_with_definition_list_after_params(param: int) -> None:  # noqa: ARG001
+def docstring_with_definition_list_after_params(param: int) -> None:
     """Do something.
 
     Args:
@@ -1132,7 +1210,7 @@ def docstring_with_definition_list_after_params(param: int) -> None:  # noqa: AR
 
     """,
 )
-def docstring_with_enum_list_after_params(param: int) -> None:  # noqa: ARG001
+def docstring_with_enum_list_after_params(param: int) -> None:
     """Do something.
 
     Args:
@@ -1143,7 +1221,7 @@ def docstring_with_enum_list_after_params(param: int) -> None:  # noqa: ARG001
     """
 
 
-@warns("Definition list ends without a blank line")
+@warns(WarningInfo(regexp="Definition list ends without a blank line", type="docutils"))
 @expected(
     """
     mod.docstring_with_definition_list_after_params_no_blank_line(param)
@@ -1167,7 +1245,7 @@ def docstring_with_enum_list_after_params(param: int) -> None:  # noqa: ARG001
        -[ Example ]-
     """,
 )
-def docstring_with_definition_list_after_params_no_blank_line(param: int) -> None:  # noqa: ARG001
+def docstring_with_definition_list_after_params_no_blank_line(param: int) -> None:
     """Do something.
 
     Args:
@@ -1240,8 +1318,6 @@ AUTO_EXCEPTION = """\
    :members:
 """
 
-LT_PY310 = sys.version_info < (3, 10)
-
 
 @expected(
     """
@@ -1268,6 +1344,19 @@ def typehints_use_signature(a: AsyncGenerator) -> AsyncGenerator:
     return a
 
 
+@expected(
+    """
+    mod.typehints_no_rtype_none()
+
+       Do something.
+
+    """,
+    typehints_document_rtype_none=False,
+)
+def typehints_no_rtype_none() -> None:
+    """Do something."""
+
+
 prolog = """
 .. |test_node_start| replace:: {test_node_start}
 """.format(test_node_start="test_start")
@@ -1292,7 +1381,7 @@ prolog = """
     """,
     rst_prolog=prolog,
 )
-def docstring_with_multiline_note_after_params_prolog_replace(param: int) -> None:  # noqa: ARG001
+def docstring_with_multiline_note_after_params_prolog_replace(param: int) -> None:
     """Do something.
 
     Args:
@@ -1329,7 +1418,7 @@ epilog = """
     """,
     rst_epilog=epilog,
 )
-def docstring_with_multiline_note_after_params_epilog_replace(param: int) -> None:  # noqa: ARG001
+def docstring_with_multiline_note_after_params_epilog_replace(param: int) -> None:
     """Do something.
 
     Args:
@@ -1346,20 +1435,86 @@ def docstring_with_multiline_note_after_params_epilog_replace(param: int) -> Non
     """
     mod.docstring_with_see_also()
 
-       Return type:
-          "str"
+       Test
 
        See also: more info at <https://example.com>`_.
 
+       Return type:
+          "str"
+
     """
 )
 def docstring_with_see_also() -> str:
     """
+    Test
+
     .. seealso:: more info at <https://example.com>`_.
     """
     return ""
 
 
+@expected(
+    """
+    mod.has_doctest1()
+
+       Test that we place the return type correctly when the function has
+       a doctest.
+
+       Return type:
+          "None"
+
+       >>> this is a fake doctest
+       a
+       >>> more doctest
+       b
+    """
+)
+def has_doctest1() -> None:
+    r"""Test that we place the return type correctly when the function has a doctest.
+
+    >>> this is a fake doctest
+    a
+    >>> more doctest
+    b
+    """
+
+
+Unformatted = TypeVar("Unformatted")
+
+
+@warns(WarningInfo(regexp="cannot cache unpickleable configuration value: 'typehints_formatter'", type="config.cache"))
+@expected(
+    """
+    mod.typehints_formatter_applied_to_signature(param: Formatted) -> Formatted
+
+       Do nothing
+
+       Parameters:
+          **param** (Formatted) -- A parameter
+
+       Return type:
+          Formatted
+
+       Returns:
+          The return value
+    """,
+    typehints_use_signature=True,
+    typehints_use_signature_return=True,
+    typehints_formatter=lambda _, __=None: "Formatted",
+)
+def typehints_formatter_applied_to_signature(param: Unformatted) -> Unformatted:
+    """
+    Do nothing
+
+    Args:
+        param: A parameter
+
+    Returns:
+        The return value
+    """
+    return param
+
+
 # Config settings for each test run.
 # Config Name: Sphinx Options as Dict.
 configs = {
@@ -1391,23 +1546,21 @@ def test_integration(
     (Path(app.srcdir) / "index.rst").write_text(template.format(val.__name__))
     app.config.__dict__.update(configs[conf_run])
     app.config.__dict__.update(val.OPTIONS)
+    app.config.always_use_bars_union = True
     monkeypatch.setitem(sys.modules, "mod", sys.modules[__name__])
     app.build()
     assert "build succeeded" in status.getvalue()  # Build succeeded
 
-    regexp = getattr(val, "WARNING", None)
+    warning_info: Union[WarningInfo, None] = getattr(val, "WARNING", None)
     value = warning.getvalue().strip()
-    if regexp:
-        msg = f"Regex pattern did not match.\n Regex: {regexp!r}\n Input: {value!r}"
-        assert re.search(regexp, value), msg
+    if warning_info:
+        warning_info.assert_warning(value)
     else:
         assert not value
 
     result = (Path(app.srcdir) / "_build/text/index.txt").read_text()
 
     expected = val.EXPECTED
-    if LT_PY310:
-        expected = expected.replace("NewType", "NewType()")
     try:
         assert result.strip() == dedent(expected).strip()
     except Exception:
